package se.umu.wsag.impl; 

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.StringWriter;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.ggf.ws_agreement.AgreementPropertiesType;
import org.ggf.ws_agreement.ServiceDescriptionTermType;
import org.ggf.ws_agreement.TermCompositorType;
import org.ggf.ws_agreement.TermTreeType;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.globus.wsrf.impl.PersistentReflectionResource;
import org.globus.wsrf.utils.FilePersistenceHelper;
import org.globus.wsrf.impl.SimpleResourceKey; 
import org.xml.sax.InputSource;

import se.umu.reservation.DecisionMakerConfig;

/**
 * The AgreementResource represents a created agreement. 
 * The RPs of this resource includes any information about the 
 * local agreement. Destruction of this resource cancels 
 * the agreement. 
 * 
 * The agreementresource is persisted to two files, one containing
 * the decisionMaker (resourkey.conf) and the other one (resourcekey.xml)
 * containing the resource itself. The .xml file is serialized using
 * the default serialization mechanism in PersistentReflectionResource, 
 * while the other file (resourcekey.conf) is serialized manually.
 * 
 */

public class AgreementResource extends PersistentReflectionResource {

	private static Logger logger = Logger.getLogger(AgreementResource.class);
	private AgreementDecisionMaker decisionMaker;		
	
	
	
	/**
	 * Loads the stored decisionMakerconfig from the persistent file, 
	 * creates a new AgreementDecisionMaker and initializes it from file. 
	 */
    public synchronized void load(ResourceKey key) throws ResourceException {
    	/*
		int keyValue = key.getValue().toString().hashCode(); 
		if(keyValue < 0) 
			keyValue *= -1; 
		String fileKey = new Integer(keyValue).toString(); 
		
		ResourceKey newKey = new SimpleResourceKey(key.getName(), fileKey); 
  
    	super.load(newKey); // Must call this, to store resource itself...
    	*/
    	super.load(key); 
    	String fileKey = key.getValue().toString(); 
    	    	    	    	    	    	
		try {
			String configFilePath = getDecisionMakerConfigPath(fileKey); 		
			logger.debug("Config file is: " + configFilePath); 
			
			// Deserialize decision maker from config file. 
			DecisionMakerConfig config = null; 
			QName configQName = DecisionMakerConfig.getTypeDesc().getXmlType(); 
			FileInputStream input = new FileInputStream(configFilePath); 
			InputSource inputSource = new InputSource(input); 
			config = (DecisionMakerConfig) ObjectDeserializer.deserialize(
					inputSource, DecisionMakerConfig.class); 
			
			// Init decisionMaker from config: 
			String decisionClassName = config.getClassName();     	    	
			this.decisionMaker = (AgreementDecisionMaker) Class.forName
			(decisionClassName).newInstance();	
			this.decisionMaker.init(config); 
			logger.debug("DecisionMaker is: " +	decisionMaker.getClass());
    	
		} catch (Exception e) {
			logger.error("Error loading resource: " + e.getMessage()); 
		}
    	
    	
    }
    
    /**
     * Performs specific persistence handling, stores configuration 
     * of AgreementDecisionMaker (DecisionMakerConfig), ensuring
     * that the decisionMaker can be recreated upon restart. 
     */
    public synchronized void store() throws ResourceException {    	    
    	
    	super.store(); 
    	
    	try {
    		// Get serialized form of AgreementDecisionMaker 
    		// (actually, of, DecisionMakerConfig)
    		// and store it in resourcekey.conf
    		DecisionMakerConfig config = this.decisionMaker.getConfig(); 
    		
    		// Write to temporary file, delete old config and rename
    		// temporary file. 
    		
    		// Create tmp file. 
    		FilePersistenceHelper helper = new FilePersistenceHelper(
    				AgreementPropertiesType.class, ".ser"); 
    		File tmpFile = File.createTempFile("decisionMakerConfig", ".tmp", 
    				helper.getStorageDirectory());     		
    		
    		// Serialize config and write it to tmp file
    		StringWriter sw = new StringWriter(); 
    		ObjectSerializer.serialize(sw, config, 
    				DecisionMakerConfig.getTypeDesc().getXmlType()); 
    		String s = sw.toString();     		    		
    		FileOutputStream fos = new FileOutputStream(tmpFile); 
    		fos.write(s.getBytes()); 
    		fos.close();     		    		
    		
    		
    		// Reverse sign in hashcode if it is negative. 
    		/*
    		int ID = this.getID().hashCode(); 
    		if(ID < 0)
    			ID *= -1; 
    		String key = new Integer(ID).toString(); 
    		*/
    		String configFilePath = getDecisionMakerConfigPath(this.getID()); 
    		// String configFilePath = getDecisionMakerConfigPath(key); 
    		File configFile = new File(configFilePath); 
    		
    		logger.debug("Storing resource config in " + configFilePath); 
    		
    		// Delete old file
    		if(configFile.exists()) {
    			configFile.delete(); 
    		}
    		if(!(tmpFile.renameTo(configFile))) {
    			logger.error("Can not move " + tmpFile.getAbsolutePath() 
    					+ " to " + configFile.getAbsolutePath()); 
    			throw new IOException("Can not write " 
    					+ configFile.getAbsolutePath()); 
    		}
    	} catch (IOException e) {
    		logger.error("Ooops: " + e.getMessage()); 
    		throw new ResourceException(e); 
    	}
    	catch (Exception e) {
    		logger.error("Error serializing decisionMakerconfig: " 
    				+ e.getMessage()); 
    		throw new ResourceException(e); 
    	}
    }
    
	// TODO: Pass whole agreement to decisionMaker? 
	// It may be useful for other applications than reservations
	/**
	 * Calls decisionMaker.cancel which cleans up any local stored
	 * information about the agreement. 
	 * Also cleans up the decisionmaker config file. 
	 */
	public void remove() {
		
		try {
			super.remove(); 
		} catch (Exception e) {
			logger.warn(e.getMessage()); 
		}
		
		try {
			logger.debug("Destroying resource with key " + this.getID());
			AgreementPropertiesType properties = (AgreementPropertiesType)
				this.getResourceBean(); 
			TermCompositorType allTerms = properties.getTerms().getAll(); 
			ServiceDescriptionTermType serviceTerms = 
				allTerms.getServiceDescriptionTerm(); 	
			this.decisionMaker.cancel(serviceTerms); 
			
			// Remove config file: 
			String configPath = this.getDecisionMakerConfigPath(this.getID()); 
			logger.debug("Removing config file: " + configPath); 
			File f = new File(configPath); 
			if(f.exists()) {
				f.delete();
			}
		} catch (Exception e) {
			logger.warn("Could not remove local agreement: " + e.getMessage());
		}
	
	}  
    

    
    
    /**
     * Gets the path to the persisted decisionMaker config. 
     * @param key Either a resourcekey or an Integer (value of resource key). 
     * @return The absolute path to the config file. 
     * @throws IOException
     */
    private String getDecisionMakerConfigPath(Object key) 
    		throws IOException {
    	
    	String fileName = null; 
    	if(key instanceof ResourceKey) {
			fileName = ((ResourceKey)key).getValue().toString(); 
    	}
    	else {
    		fileName = key.toString(); 
    	}
    	
		// Get config file (resourcekey.conf)
		FilePersistenceHelper helper = 
			new FilePersistenceHelper(AgreementPropertiesType.class, ".conf"); 
		File f = helper.getStorageDirectory(); 		
		String configFilePath = f.getAbsolutePath() 
				+ "/"
				+ fileName
				+ helper.getFileSuffix(); 
		logger.debug("Config file is: " + configFilePath); 
		
		return configFilePath; 
    }
    

	
	/**
	 * @param decisionMaker The decisionMaker to set.
	 */
	public void setDecisionMaker(AgreementDecisionMaker decMaker) {
		logger.debug("Setting decionMaker"); 
		this.decisionMaker = decMaker; 
	}
	
	// Specified by PersistentReflectionResource
	public Class getResourceBeanClass() {		
		return AgreementPropertiesType.class; 
	}
		
}
